from typing import TYPE_CHECKING, Optional, Tuple, cast

from passlib.context import CryptContext

if TYPE_CHECKING:
    from pydantic import SecretStr


class PasswordManager:
    """Thin wrapper around `passlib`."""

    def __init__(self) -> None:
        """Construct a PasswordManager."""

        self.context = CryptContext(schemes=["bcrypt"], deprecated="auto")

    def get_hash(self, password: "SecretStr") -> str:
        """Create a password hash.

        Args:
            password: The password to hash.
        """
        return cast("str", self.context.hash(password.get_secret_value()))

    def verify_and_update(self, password: "SecretStr", password_hash: str) -> Tuple[bool, Optional[str]]:
        """Verify a password and rehash it if the hash is deprecated.

        Args:
            password: The password to verify.
            password_hash: The hash to verify against.
        """
        return cast(
            "Tuple[bool, Optional[str]]", self.context.verify_and_update(password.get_secret_value(), password_hash)
        )



from typing import TYPE_CHECKING, Type

from starlite.exceptions import (
    HTTPException,
    InternalServerException,
    NotFoundException,
)
from starlite.middleware.exceptions.debug_response import create_debug_response
from starlite.utils.exception import create_exception_response

if TYPE_CHECKING:
    from starlite import Request, Response


class RepositoryException(Exception):
    """Base repository exception."""


class RepositoryNotFoundException(RepositoryException):
    """Raise when a db record is expected but none is found."""


class RepositoryConflictException(RepositoryException):
    """Raise when db constraints are violated."""


class TokenException(Exception):
    """Base token exception."""


class InvalidTokenException(TokenException):
    """Raise when a JWT is found to be invalid."""


class ExpiredTokenException(TokenException):
    """Raise when a JWT is found to be expired."""


class ConflictException(HTTPException):
    """Generic HTTP conflict exception."""

    status_code = 409


class InvalidException(HTTPException):
    """Generic HTTP invalid exception."""

    status_code = 400


def _create_exception_response(
    request: "Request", exception: Exception, http_exception: Type[HTTPException]
) -> "Response":
    """Create an appropriate response depending on `request.app.debug` value."""
    if http_exception is InternalServerException and request.app.debug:
        return create_debug_response(request, exception)
    return create_exception_response(http_exception())


def repository_exception_handler(request: "Request", exception: RepositoryException) -> "Response":
    """Transform repository exceptions to HTTP exceptions."""

    http_exception: type[HTTPException]
    if isinstance(exception, RepositoryNotFoundException):
        http_exception = NotFoundException
    elif isinstance(exception, RepositoryConflictException):
        http_exception = ConflictException
    else:
        http_exception = InternalServerException
    return _create_exception_response(request, exception, http_exception)


def token_exception_handler(request: "Request", exception: TokenException) -> "Response":
    """Transform token exceptions to HTTP exceptions."""

    http_exception: type[HTTPException]
    if isinstance(exception, (InvalidTokenException, ExpiredTokenException)):
        http_exception = InvalidException
    else:
        http_exception = InternalServerException
    return _create_exception_response(request, exception, http_exception)



from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Any, Generic, Optional, Type, TypeVar

from jose import JWTError
from pydantic import SecretStr
from starlite.contrib.jwt.jwt_token import Token
from starlite.exceptions import ImproperlyConfiguredException

from starlite_users.adapter.sqlalchemy.mixins import (
    RoleModelType,
    UserModelType,
    UserRoleModelType,
)
from starlite_users.adapter.sqlalchemy.repository import (
    SQLAlchemyUserRepository,
    SQLAlchemyUserRoleRepository,
)
from starlite_users.exceptions import (
    InvalidTokenException,
    RepositoryConflictException,
    RepositoryNotFoundException,
)
from starlite_users.password import PasswordManager
from starlite_users.schema import (
    RoleCreateDTOType,
    RoleUpdateDTOType,
    UserAuthSchema,
    UserCreateDTOType,
    UserUpdateDTOType,
)

if TYPE_CHECKING:
    from uuid import UUID


class BaseUserService(Generic[UserModelType, UserCreateDTOType, UserUpdateDTOType]):  # pylint: disable=R0904
    """Main user management interface."""

    user_model: Type[UserModelType]
    """A subclass of a `User` ORM model."""
    secret: SecretStr
    """Secret string for securely signing tokens."""

    def __init__(self, repository: "SQLAlchemyUserRepository[UserModelType]") -> None:
        """User service constructor.

        Args:
            repository: A `UserRepository` instance
        """
        self.repository = repository
        self.password_manager = PasswordManager()

    async def add_user(self, data: UserCreateDTOType, process_unsafe_fields: bool = False) -> UserModelType:
        """Create a new user programmatically.

        Args:
            data: User creation data transfer object.
            process_unsafe_fields: If True, set `is_active` and `is_verified` attributes as they appear in `data`, otherwise always set their defaults.
        """
        try:
            existing_user = await self.get_user_by(email=data.email)
            if existing_user:
                raise RepositoryConflictException("email already associated with an account")
        except RepositoryNotFoundException:
            pass

        user_dict = data.dict(exclude={"password"})
        user_dict["password_hash"] = self.password_manager.get_hash(data.password)
        if not process_unsafe_fields:
            user_dict["is_verified"] = False
            user_dict["is_active"] = True

        return await self.repository.add_user(self.user_model(**user_dict))  # type: ignore[arg-type]

    async def register(self, data: UserCreateDTOType) -> Optional[UserModelType]:
        """Register a new user and optionally run custom business logic.

        Args:
            data: User creation data transfer object.
        """
        if not await self.pre_registration_hook(data):
            return None

        user = await self.add_user(data)
        await self.initiate_verification(user)  # TODO: make verification optional?

        await self.post_registration_hook(user)

        return user

    async def get_user(self, id_: "UUID") -> UserModelType:
        """Retrieve a user from the database by id.

        Args:
            id_: UUID corresponding to a user primary key.
        """
        return await self.repository.get_user(id_)

    async def get_user_by(self, **kwargs: Any) -> UserModelType:
        """Retrieve a user from the database by arbitrary keyword arguments.

        Args:
            **kwargs: Keyword arguments to pass as filters.

        Examples:
            ```python
            service = UserService(...)
            john = await service.get_user_by(email="john@example.com")
            ```
        """
        return await self.repository.get_user_by(**kwargs)

    async def update_user(self, id_: "UUID", data: UserUpdateDTOType) -> UserModelType:
        """Update arbitrary user attributes in the database.

        Args:
            id_: UUID corresponding to a user primary key.
            data: User update data transfer object.
        """
        update_dict = data.dict(exclude={"password"}, exclude_unset=True)
        if data.password:
            update_dict["password_hash"] = self.password_manager.get_hash(data.password)

        return await self.repository.update_user(id_, update_dict)

    async def delete_user(self, id_: "UUID") -> None:
        """Delete a user from the database.

        Args:
            id_: UUID corresponding to a user primary key.
        """
        return await self.repository.delete_user(id_)

    async def authenticate(self, data: UserAuthSchema) -> Optional[UserModelType]:
        """Authenticate a user.

        Args:
            data: User authentication data transfer object.
        """
        if not await self.pre_login_hook(data):
            return None

        user = await self.repository.get_user_by(email=data.email)

        verified, new_password_hash = self.password_manager.verify_and_update(data.password, user.password_hash)
        if not verified:
            return None
        if new_password_hash is not None:
            user = await self.repository._update(user, {"password_hash": new_password_hash})

        await self.post_login_hook(user)

        return user

    def generate_token(self, user_id: "UUID", aud: str) -> str:
        """Generate a limited time valid JWT.

        Args:
            user_id: UUID of the user to provide the token to.
            aud: Context of the token
        """
        token = Token(
            exp=datetime.now() + timedelta(seconds=60 * 60 * 24),  # TODO: make time configurable?
            sub=str(user_id),
            aud=aud,
        )
        return token.encode(secret=self.secret.get_secret_value(), algorithm="HS256")

    async def initiate_verification(self, user: UserModelType) -> None:
        """Initiate the user verification flow.

        Args:
            user: The user requesting verification.
        """
        token = self.generate_token(user.id, aud="verify")
        await self.send_verifification_token(user, token)

    async def send_verifification_token(self, user: UserModelType, token: str) -> None:
        """Execute custom logic to send the verification token to the relevant user.

        Args:
            user: The user requesting verification.
            token: An encoded JWT bound to verification.

        Notes:
        - Develepors need to override this method to facilitate sending the token via email, sms etc.
        """

    async def verify(self, encoded_token: str) -> UserModelType:
        """Verify a user with the given JWT.

        Args:
            encoded_token: An encoded JWT bound to verification.

        Raises:
            InvalidTokenException: If the token is expired or tampered with.
        """
        token = self._decode_and_verify_token(encoded_token, context="verify")

        user_id = token.sub
        try:
            user = await self.repository.update_user(user_id, {"is_verified": True})
        except RepositoryNotFoundException as e:
            raise InvalidTokenException from e

        await self.post_verification_hook(user)

        return user

    async def initiate_password_reset(self, email: str) -> None:
        """Initiate the password reset flow.

        Args:
            email: Email of the user who has forgotten their password.
        """
        try:
            user = await self.get_user_by(email=email)  # TODO: something about timing attacks.
        except RepositoryNotFoundException:
            return
        token = self.generate_token(user.id, aud="reset_password")
        await self.send_password_reset_token(user, token)

    async def send_password_reset_token(self, user: UserModelType, token: str) -> None:
        """Execute custom logic to send the password reset token to the relevant user.

        Args:
            user: The user requesting the password reset.
            token: An encoded JWT bound to the password reset flow.

        Notes:
        - Develepors need to override this method to facilitate sending the token via email, sms etc.
        """

    async def reset_password(self, encoded_token: str, password: SecretStr) -> None:
        """Reset a user's password given a valid JWT.

        Args:
            encoded_token: An encoded JWT bound to the password reset flow.
            password: The new password to hash and store.

        Raises:
            InvalidTokenException: If the token has expired or been tampered with.
        """
        token = self._decode_and_verify_token(encoded_token, context="reset_password")

        user_id = token.sub
        try:
            await self.repository.update_user(user_id, {"password_hash": self.password_manager.get_hash(password)})
        except RepositoryNotFoundException as e:
            raise InvalidTokenException from e

    async def pre_login_hook(self, data: UserAuthSchema) -> bool:  # pylint: disable=W0613
        """Execute custom logic to run custom business logic prior to authenticating a user.

        Useful for authorization checks against external sources,
        eg. current membership validity or blacklists, etc
        Must return `False` or raise a custom exception to cancel authentication.

        Args:
            data: Authentication data transfer object.

        Returns:
            True: If authentication should proceed
            False: If authentication is not to proceed.

        Notes:
            Uncaught exceptions in this method will break the authentication process.
        """

        return True

    async def post_login_hook(self, user: UserModelType) -> None:
        """Execute custom logic to run custom business logic after authenticating a user.

        Useful for eg. updating a login counter, updating last known user IP
        address, etc.

        Args:
            user: The user who has authenticated.

        Notes:
            Uncaught exceptions in this method will break the authentication process.
        """

    async def pre_registration_hook(self, data: UserCreateDTOType) -> bool:  # pylint: disable=W0613
        """Execute custom logic to run custom business logic prior to registering a user.

        Useful for authorization checks against external sources,
        eg. membership API or blacklists, etc.
        Must return `False` or raise a custom exception to cancel registration.

        Args:
            data: User creation data transfer object

        Returns:
            True: If registration should proceed
            False: If registration is not to proceed.

        Notes:
        - Uncaught exceptions in this method will result in failed registration attempts.
        """

        return True

    async def post_registration_hook(self, user: UserModelType) -> None:
        """Execute custom logic to run custom business logic after registering a user.

        Useful for updating external datasets, sending welcome messages etc.

        Args:
            user: User ORM instance.

        Notes:
        - Uncaught exceptions in this method could result in returning a HTTP 500 status
        code while successfully creating the user in the database.
        - It's possible to skip verification entirely by setting `user.is_verified`
        to `True` here.
        """

    async def post_verification_hook(self, user: UserModelType) -> None:
        """Execute custom logic to run custom business logic after a user has verified details.

        Useful for eg. updating sales lead data, etc.

        Args:
            user: User ORM instance.

        Notes:
        - Uncaught exceptions in this method could result in returning a HTTP 500 status
        code while successfully validating the user.
        """

    def _decode_and_verify_token(self, encoded_token: str, context: str) -> Token:
        try:
            token = Token.decode(
                encoded_token=encoded_token,
                secret=self.secret.get_secret_value(),
                algorithm="HS256",
            )
        except JWTError as e:
            raise InvalidTokenException from e

        if token.aud != context:
            raise InvalidTokenException(f"aud value must be {context}")

        return token


class BaseUserRoleService(
    BaseUserService, Generic[UserRoleModelType, UserCreateDTOType, UserUpdateDTOType, RoleModelType]
):
    """Main user and role management interface."""

    role_model: Type[RoleModelType]
    """A subclass of a `Role` ORM model."""
    repository: "SQLAlchemyUserRoleRepository[UserRoleModelType, RoleModelType]"

    def __init__(self, repository: "SQLAlchemyUserRoleRepository[UserRoleModelType, RoleModelType]") -> None:
        """User service constructor.

        Args:
            repository: A `UserRepository` instance
        """
        self.repository = repository
        self.password_manager = PasswordManager()

    async def get_role(self, id_: "UUID") -> RoleModelType:
        """Retrieve a role by id.

        Args:
            id_: UUID of the role.
        """
        return await self.repository.get_role(id_)

    async def get_role_by_name(self, name: str) -> RoleModelType:
        """Retrieve a role by name.

        Args:
            name: The name of the role.
        """
        return await self.repository.get_role_by_name(name)

    async def add_role(self, data: RoleCreateDTOType) -> RoleModelType:
        """Add a new role to the database.

        Args:
            data: A role creation data transfer object.
        """
        return await self.repository.add_role(self.role_model(**data.dict()))

    async def update_role(self, id_: "UUID", data: RoleUpdateDTOType) -> RoleModelType:
        """Update a role in the database.

        Args:
            id_: UUID corresponding to the role primary key.
            data: A role update data transfer object.
        """
        return await self.repository.update_role(id_, data.dict(exclude_unset=True))

    async def delete_role(self, id_: "UUID") -> None:
        """Delete a role from the database.

        Args:
            id_: UUID corresponding to the role primary key.
        """
        return await self.repository.delete_role(id_)

    async def assign_role_to_user(self, user_id: "UUID", role_id: "UUID") -> UserRoleModelType:
        """Add a role to a user.

        Args:
            user_id: UUID of the user to receive the role.
            role_id: UUID of the role to add to the user.
        """
        user = await self.get_user(user_id)
        role = await self.get_role(role_id)
        if not hasattr(user, "roles"):
            raise ImproperlyConfiguredException("'User' model has no roles attribute")
        if isinstance(user.roles, list) and role in user.roles:
            raise RepositoryConflictException(f"user already has role '{role.name}'")
        return await self.repository.assign_role_to_user(user, role)

    async def revoke_role_from_user(self, user_id: "UUID", role_id: "UUID") -> UserRoleModelType:
        """Revoke a role from a user.

        Args:
            user_id: UUID of the user to revoke the role from.
            role_id: UUID of the role to revoke.
        """
        user = await self.get_user(user_id)
        role = await self.get_role(role_id)
        if not hasattr(user, "roles"):
            raise ImproperlyConfiguredException("'User' model has no roles attribute")
        if isinstance(user.roles, list) and role not in user.roles:
            raise RepositoryConflictException(f"user does not have role '{role.name}'")
        return await self.repository.revoke_role_from_user(user, role)


UserServiceType = TypeVar("UserServiceType", bound=BaseUserService)
UserRoleServiceType = TypeVar("UserRoleServiceType", bound=BaseUserRoleService)



from typing import Any, Dict, Generic, List, Literal, Optional, Type

from pydantic import BaseModel, SecretStr, root_validator
from starlite.middleware.session.base import BaseBackendConfig
from starlite.types import Guard

from starlite_users.adapter.sqlalchemy.mixins import RoleModelType, UserModelType
from starlite_users.schema import (
    RoleCreateDTOType,
    RoleReadDTOType,
    RoleUpdateDTOType,
    UserCreateDTOType,
    UserReadDTOType,
    UserUpdateDTOType,
)
from starlite_users.service import UserServiceType


class AuthHandlerConfig(BaseModel):
    """Configuration for user authentication route handlers.

    Passing an instance to `StarliteUsersConfig` will automatically take care of handler registration on the app.
    """

    login_path: str = "/login"
    """The path for the user authentication/login route."""
    logout_path: str = "/logout"
    """The path for the logout route."""


class CurrentUserHandlerConfig(BaseModel):
    """Configuration for the current-user route handler.

    Passing an instance to `StarliteUsersConfig` will automatically take care of handler registration on the app.
    """

    path: str = "/users/me"
    """The path to get or update the currently logged-in user."""


class PasswordResetHandlerConfig(BaseModel):
    """Configuration for the forgot-password and reset-password route handlers.

    Passing an instance to `StarliteUsersConfig` will automatically take care of handler registration on the app.
    """

    forgot_path: str = "/forgot-password"
    """The path for the forgot-password route."""
    reset_path: str = "/reset-password"
    """The path for the reset-password route."""


class RegisterHandlerConfig(BaseModel):
    """Configuration for the user registration route handler.

    Passing an instance to `StarliteUsersConfig` will automatically take care of handler registration on the app.
    """

    path: str = "/register"
    """The path for the registration/signup route."""


class RoleManagementHandlerConfig(BaseModel):
    """Configuration for the role management route handlers.

    Passing an instance to `StarliteUsersConfig` will automatically take care of handler registration on the app.
    """

    path_prefix: str = "/users/roles"
    """The prefix for the router path."""
    assign_role_path: str = "/assign"
    """The path for the role assignment router."""
    revoke_role_path: str = "/revoke"
    """The path for the role revokement router."""
    guards: List[Guard]
    """A list of callable [Guards][starlite.types.Guard] that determines who is authorized to manage roles."""
    opt: Dict[str, Any] = {}
    """Optional route handler 'opts' to provide additional context to Guards.

    Note:
        - See https://starlite-api.github.io/starlite/1.48/usage/8-security/3-guards/#the-route-handler-opt-key for more info.
    """


class UserManagementHandlerConfig(BaseModel):
    """Configuration for user management (read, update, delete) route handlers.

    Passing an instance to `StarliteUsersConfig` will automatically take care of handler registration on the app.

    Note:
    - These routes make use of Starlite `Guard`s to require authorization. Callers require admin or similar privileges.
    """

    path_prefix: str = "/users"
    """The prefix for the router path.

    By default, the path will be suffixed with `'/{id_:uuid}'`.
    """
    guards: List[Guard]
    """A list of callable [Guards][starlite.types.Guard] that determines who is authorized to manage other users."""
    opt: Dict[str, Any] = {}
    """"""


class VerificationHandlerConfig(BaseModel):
    """Configuration for the user verification route handler.

    Passing an instance to `StarliteUsersConfig` will automatically take care of handler registration on the app.
    """

    path: str = "/verify"
    """The path for the verification route."""


class StarliteUsersConfig(
    BaseModel,
    Generic[
        UserModelType,
        UserCreateDTOType,
        UserServiceType,
        UserReadDTOType,
        UserUpdateDTOType,
        RoleModelType,
        RoleCreateDTOType,
        RoleReadDTOType,
        RoleUpdateDTOType,
    ],
):
    """Configuration class for StarliteUsers."""

    class Config:
        arbitrary_types_allowed = True

    auth_exclude_paths: List[str] = ["/schema"]
    """Paths to be excluded from authentication checks."""
    auth_backend: Literal["session", "jwt", "jwt_cookie"]
    """The authentication backend to use by Starlite."""
    secret: SecretStr
    """Secret string for securely signing tokens."""
    session_backend_config: Optional[BaseBackendConfig] = None
    """Optional backend configuration for session based authentication.

    Notes:
        - Required if `auth_backend` is set to `session`.
    """
    user_model: Type[UserModelType]
    """A subclass of a `User` ORM model."""
    user_create_dto: Type[UserCreateDTOType]
    """A subclass of [BaseUserCreateDTO][starlite_users.schema.BaseUserCreateDTO]."""
    user_read_dto: Type[UserReadDTOType]
    """A subclass of [BaseUserReadDTO][starlite_users.schema.BaseUserReadDTO]."""
    user_update_dto: Type[UserUpdateDTOType]
    """A subclass of [BaseUserUpdateDTO][starlite_users.schema.BaseUserUpdateDTO]."""
    role_model: Optional[Type[RoleModelType]] = None
    """A subclass of a `Role` ORM model.

    Notes:
        - Required if `role_management_handler_config` is set.
    """
    role_create_dto: Optional[Type[RoleCreateDTOType]]
    """A subclass of [BaseRoleCreateDTO][starlite_users.schema.BaseRoleCreateDTO].

    Notes:
        - Required if `role_management_handler_config` is set.
    """
    role_read_dto: Optional[Type[RoleReadDTOType]]
    """A subclass of [BaseRoleReadDTO][starlite_users.schema.BaseRoleReadDTO].

    Notes:
        - Required if `role_management_handler_config` is set.
    """
    role_update_dto: Optional[Type[RoleUpdateDTOType]]
    """A subclass of [BaseRoleUpdateDTO][starlite_users.schema.BaseRoleUpdateDTO].

    Notes:
        - Required if `role_management_handler_config` is set.
    """
    user_service_class: Type[UserServiceType]
    """A subclass of [BaseUserService][starlite_users.service.BaseUserService]."""
    auth_handler_config: Optional[AuthHandlerConfig]
    """Optional instance of [AuthHandlerConfig][starlite_users.config.AuthHandlerConfig]. If set, registers the route
    handler(s) on the app.

    Note:
        - At least one route handler config must be set.
    """
    current_user_handler_config: Optional[CurrentUserHandlerConfig]
    """Optional current-user route handler configuration. If set, registers the route handler(s) on the app.

    Note:
        - At least one route handler config must be set.
    """
    password_reset_handler_config: Optional[PasswordResetHandlerConfig]
    """Optional password reset route handler configuration. If set, registers the route handler(s) on the app.

    Note:
        - At least one route handler config must be set.
    """
    register_handler_config: Optional[RegisterHandlerConfig]
    """Optional registration/signup route handler configuration. If set, registers the route handler(s) on the app.

    Note:
        - At least one route handler config must be set.
    """
    role_management_handler_config: Optional[RoleManagementHandlerConfig]
    """Optional role management route handler configuration. If set, registers the route handler(s) on the app.

    Note:
        - At least one route handler config must be set.
    """
    user_management_handler_config: Optional[UserManagementHandlerConfig]
    """Optional user management route handler configuration. If set, registers the route handler(s) on the app.

    Note:
        - At least one route handler config must be set.
    """
    verification_handler_config: Optional[VerificationHandlerConfig]
    """Optional user verification route handler configuration. If set, registers the route handler(s) on the app.

    Note:
        - At least one route handler config must be set.
    """

    @root_validator
    def validate_config(cls, values: Dict[str, Any]) -> Dict[str, Any]:  # pylint: disable=E0213
        """Validate the configuration.

        - A session backend must be configured if `auth_backend` is set to `'session'`.
        - At least one route handler must be configured.
        - `role_model`, `role_create_dto`, `role_read_dto` and `role_update_dto` are required fields if `role_management_handler_config` is configured.
        """
        if values.get("auth_backend") == "session" and not values.get("session_backend_config"):
            raise ValueError('session_backend_config must be set when auth_backend is set to "session"')
        handler_configs = [
            "auth_handler_config",
            "current_user_handler_config",
            "password_reset_handler_config",
            "register_handler_config",
            "role_management_handler_config",
            "user_management_handler_config",
            "verification_handler_config",
        ]
        if all(values.get(config) is None for config in handler_configs):
            raise ValueError("at least one route handler must be configured")
        if values.get("role_management_handler_config"):
            if values.get("role_model") is None:
                raise ValueError("role_model must be set when role_management_handler_config is set")
            if values.get("role_create_dto") is None:
                raise ValueError("role_create_dto must be set when role_management_handler_config is set")
            if values.get("role_read_dto") is None:
                raise ValueError("role_read_dto must be set when role_management_handler_config is set")
            if values.get("role_update_dto") is None:
                raise ValueError("role_update_dto must be set when role_management_handler_config is set")

        return values



from typing import TYPE_CHECKING, Callable, Optional, Type

from sqlalchemy.orm import Session
from starlite import State  # noqa: TC002
from starlite.exceptions import ImproperlyConfiguredException
from starlite.plugins.sql_alchemy import SQLAlchemyPlugin
from starlite.types import Scope  # noqa: TC002

from starlite_users.adapter.sqlalchemy.mixins import SQLAlchemyUserRoleMixin
from starlite_users.adapter.sqlalchemy.repository import (
    SQLAlchemyUserRepository,
    SQLAlchemyUserRoleRepository,
)

if TYPE_CHECKING:
    from starlite_users.adapter.sqlalchemy.mixins import RoleModelType, UserModelType
    from starlite_users.service import UserServiceType


def get_service_dependency(
    user_model: Type["UserModelType"],
    role_model: Optional[Type["RoleModelType"]],
    user_service_class: Type["UserServiceType"],
) -> Callable:
    """Get a service dependency callable.

    Args:
        user_model: A subclass of a `User` ORM model.
        role_model: A subclass of a `Role` ORM model.
        user_service_class: A subclass of [BaseUserService][starlite_users.service.BaseUserService]
    """

    def get_service(scope: "Scope", state: "State") -> "UserServiceType":
        """Instantiate service and repository for use with DI.

        Args:
            scope: ASGI scope
            state: The application.state instance
        """
        session = None
        if not any(isinstance(plugin, SQLAlchemyPlugin) for plugin in scope["app"].plugins):
            raise ImproperlyConfiguredException("SQLAlchemyPlugin must be configured with SQLAlchemyConfig")
        for plugin in scope["app"].plugins:
            if isinstance(plugin, SQLAlchemyPlugin):
                if plugin._config is None:
                    raise ImproperlyConfiguredException("SQLAlchemyPlugin must be configured with SQLAlchemyConfig")
                session = plugin._config.create_db_session_dependency(state, scope)
                break

        if session is None or isinstance(session, Session):
            raise ImproperlyConfiguredException("session must be instance of sqlalchemy.AsyncSession")

        if issubclass(user_model, SQLAlchemyUserRoleMixin) and role_model:
            repository = SQLAlchemyUserRoleRepository(
                session=session, user_model_type=user_model, role_model_type=role_model
            )
        else:
            repository = SQLAlchemyUserRepository(session=session, user_model_type=user_model)  # type: ignore[assignment]

        return user_service_class(repository)

    return get_service



from typing import TYPE_CHECKING, Any, Callable, Dict, List, Sequence, Type, Union

from starlite import HTTPRouteHandler, OpenAPIConfig, Request, Response, Router
from starlite.contrib.jwt import JWTAuth, JWTCookieAuth
from starlite.security.session_auth import SessionAuth

from starlite_users.dependencies import get_service_dependency
from starlite_users.exceptions import (
    RepositoryException,
    TokenException,
    repository_exception_handler,
    token_exception_handler,
)
from starlite_users.route_handlers import (
    get_auth_handler,
    get_current_user_handler,
    get_password_reset_handler,
    get_registration_handler,
    get_role_management_handler,
    get_user_management_handler,
    get_verification_handler,
)
from starlite_users.user_handlers import (
    get_jwt_retrieve_user_handler,
    get_session_retrieve_user_handler,
)

if TYPE_CHECKING:
    from starlite.config import AppConfig

    from starlite_users.config import StarliteUsersConfig

EXCLUDE_AUTH_HANDLERS = (
    "login",
    "register",
    "verify",
    "forgot_password",
    "reset_password",
)


class StarliteUsers:
    """A Starlite extension for authentication, authorization and user management."""

    def __init__(self, config: "StarliteUsersConfig") -> None:
        """Construct a StarliteUsers instance."""
        self._config = config

    def on_app_init(self, app_config: "AppConfig") -> "AppConfig":
        """Register routers, auth strategies etc on the Starlite app.

        Args:
            app_config: An instance of [AppConfig][starlite.config.AppConfig]
        """
        auth_exclude_paths = {*self._config.auth_exclude_paths}
        auth_backend = self._get_auth_backend()
        route_handlers = self._get_route_handlers(auth_backend)

        for router in route_handlers:
            if isinstance(router, Router):
                for route in router.routes:
                    if any(name in EXCLUDE_AUTH_HANDLERS for name in route.handler_names):
                        auth_exclude_paths.add(route.path)
            if isinstance(router, HTTPRouteHandler) and router.handler_name in EXCLUDE_AUTH_HANDLERS:
                auth_exclude_paths.update(router.paths)

        app_config.openapi_config = OpenAPIConfig(
            title="Security API",  # TODO: make configurable
            version="0.1.0",  # TODO: make configurable
        )
        # will always be true
        if isinstance(auth_backend.exclude, list):
            auth_backend.exclude.extend(auth_exclude_paths)
        app_config = auth_backend.on_app_init(app_config)
        app_config.route_handlers.extend(route_handlers)

        exception_handlers: Dict[Type[Exception], Callable[[Request, Any], Response]] = {
            TokenException: token_exception_handler,
            RepositoryException: repository_exception_handler,
        }
        app_config.exception_handlers.update(exception_handlers)  # type: ignore[arg-type]

        return app_config

    def _get_auth_backend(self) -> Union[JWTAuth, JWTCookieAuth, SessionAuth]:
        if self._config.auth_backend == "session":
            return SessionAuth(
                retrieve_user_handler=get_session_retrieve_user_handler(
                    self._config.user_model, self._config.role_model
                ),
                session_backend_config=self._config.session_backend_config,  # type: ignore
                exclude=[],
            )
        if self._config.auth_backend == "jwt":
            return JWTAuth(
                retrieve_user_handler=get_jwt_retrieve_user_handler(self._config.user_model, self._config.role_model),
                token_secret=self._config.secret.get_secret_value(),
                exclude=[],
            )

        return JWTCookieAuth(
            retrieve_user_handler=get_jwt_retrieve_user_handler(self._config.user_model, self._config.role_model),
            token_secret=self._config.secret.get_secret_value(),
            exclude=[],
        )

    def _get_route_handlers(
        self, auth_backend: Union[JWTAuth, JWTCookieAuth, SessionAuth]
    ) -> Sequence[Union[HTTPRouteHandler, Router]]:
        """Parse the route handler configs to get Routers."""

        handlers: List[Union[HTTPRouteHandler, Router]] = []
        if self._config.auth_handler_config:
            handlers.append(
                get_auth_handler(
                    login_path=self._config.auth_handler_config.login_path,
                    logout_path=self._config.auth_handler_config.logout_path,
                    user_read_dto=self._config.user_read_dto,
                    service_dependency=get_service_dependency(
                        self._config.user_model, self._config.role_model, self._config.user_service_class
                    ),
                    auth_backend=auth_backend,
                )
            )
        if self._config.current_user_handler_config:
            handlers.append(
                get_current_user_handler(
                    path=self._config.current_user_handler_config.path,
                    user_read_dto=self._config.user_read_dto,
                    user_update_dto=self._config.user_update_dto,
                    service_dependency=get_service_dependency(
                        self._config.user_model, self._config.role_model, self._config.user_service_class
                    ),
                )
            )
        if self._config.password_reset_handler_config:
            handlers.append(
                get_password_reset_handler(
                    forgot_path=self._config.password_reset_handler_config.forgot_path,
                    reset_path=self._config.password_reset_handler_config.reset_path,
                    service_dependency=get_service_dependency(
                        self._config.user_model, self._config.role_model, self._config.user_service_class
                    ),
                )
            )
        if self._config.register_handler_config:
            handlers.append(
                get_registration_handler(
                    path=self._config.register_handler_config.path,
                    user_create_dto=self._config.user_create_dto,
                    user_read_dto=self._config.user_read_dto,
                    service_dependency=get_service_dependency(
                        self._config.user_model, self._config.role_model, self._config.user_service_class
                    ),
                )
            )
        if self._config.role_management_handler_config:
            handlers.append(
                get_role_management_handler(
                    path_prefix=self._config.role_management_handler_config.path_prefix,
                    assign_role_path=self._config.role_management_handler_config.assign_role_path,
                    revoke_role_path=self._config.role_management_handler_config.revoke_role_path,
                    guards=self._config.role_management_handler_config.guards,
                    opt=self._config.role_management_handler_config.opt,
                    role_create_dto=self._config.role_create_dto,  # type: ignore[arg-type]
                    role_read_dto=self._config.role_read_dto,  # type: ignore[arg-type]
                    role_update_dto=self._config.role_update_dto,  # type: ignore[arg-type]
                    user_read_dto=self._config.user_read_dto,
                    service_dependency=get_service_dependency(
                        self._config.user_model, self._config.role_model, self._config.user_service_class
                    ),
                )
            )
        if self._config.user_management_handler_config:
            handlers.append(
                get_user_management_handler(
                    path_prefix=self._config.user_management_handler_config.path_prefix,
                    guards=self._config.user_management_handler_config.guards,
                    opt=self._config.user_management_handler_config.opt,
                    user_read_dto=self._config.user_read_dto,
                    user_update_dto=self._config.user_update_dto,
                    service_dependency=get_service_dependency(
                        self._config.user_model, self._config.role_model, self._config.user_service_class
                    ),
                )
            )
        if self._config.verification_handler_config:
            handlers.append(
                get_verification_handler(
                    path=self._config.verification_handler_config.path,
                    user_read_dto=self._config.user_read_dto,
                    service_dependency=get_service_dependency(
                        self._config.user_model, self._config.role_model, self._config.user_service_class
                    ),
                )
            )
        return handlers



from typing import TYPE_CHECKING, List, TypeVar, Union
from uuid import UUID, uuid4

from sqlalchemy import Boolean, Column, ForeignKey, String
from sqlalchemy.orm import relationship
from sqlalchemy.orm.decl_api import declarative_mixin, declared_attr

from starlite_users.adapter.sqlalchemy.guid import GUID

if TYPE_CHECKING:
    from sqlalchemy.orm.attributes import Mapped  # type: ignore[attr-defined]


@declarative_mixin
class SQLAlchemyUserMixin:
    """Base SQLAlchemy user mixin."""

    __tablename__ = "user"

    id: "Mapped[UUID]" = Column(
        GUID(),
        primary_key=True,
        default=uuid4,
        unique=True,
        nullable=False,
    )
    email: "Mapped[str]" = Column(String(320), nullable=False, unique=True)
    password_hash: "Mapped[str]" = Column(String(1024))
    is_active: "Mapped[bool]" = Column(Boolean(), nullable=False, default=False)
    is_verified: "Mapped[bool]" = Column(Boolean(), nullable=False, default=False)


@declarative_mixin
class SQLAlchemyUserRoleMixin(SQLAlchemyUserMixin):
    """Base SQLAlchemy user mixin with `roles` attribute."""

    __tablename__ = "user"

    @declared_attr  # type: ignore[misc]
    def roles(cls) -> "Mapped[List[RoleModelType]]":  # pylint: disable=E0213
        """Roles attribute."""

        return relationship("Role", secondary="user_roles", lazy="joined")  # type: ignore[misc]


@declarative_mixin
class SQLAlchemyRoleMixin:
    """Base SQLAlchemy role mixin."""

    __tablename__ = "role"

    id: "Mapped[UUID]" = Column(
        GUID(),
        primary_key=True,
        default=uuid4,
        unique=True,
        nullable=False,
    )
    name: "Mapped[str]" = Column(String(255), nullable=False, unique=True)
    description: "Mapped[str]" = Column(String(255), nullable=True)


@declarative_mixin
class UserRoleAssociationMixin:
    """Base SQLAlchemy `user.roles` association mixin."""

    __tablename__ = "user_roles"

    @declared_attr  # type: ignore[misc]
    def role_id(cls) -> "Mapped[UUID]":  # pylint: disable=E0213
        """Id attribute."""

        return Column(GUID(), ForeignKey("role.id"), nullable=True)

    @declared_attr  # type: ignore[misc]
    def user_id(cls) -> "Mapped[UUID]":  # pylint: disable=E0213
        """Id attribute."""

        return Column(GUID(), ForeignKey("user.id"), nullable=True)


UserModelType = TypeVar("UserModelType", bound=Union[SQLAlchemyUserMixin, SQLAlchemyUserRoleMixin])
UserRoleModelType = TypeVar("UserRoleModelType", bound=SQLAlchemyUserRoleMixin)
RoleModelType = TypeVar("RoleModelType", bound=SQLAlchemyRoleMixin)



from typing import TYPE_CHECKING, Any, Callable, Dict, Optional, Type

from starlite_users.adapter.sqlalchemy.mixins import SQLAlchemyUserRoleMixin
from starlite_users.adapter.sqlalchemy.repository import (
    SQLAlchemyUserRepository,
    SQLAlchemyUserRoleRepository,
)
from starlite_users.exceptions import RepositoryNotFoundException

if TYPE_CHECKING:
    from starlite import ASGIConnection
    from starlite.contrib.jwt import Token

    from starlite_users.adapter.sqlalchemy.mixins import RoleModelType, UserModelType


def get_session_retrieve_user_handler(
    user_model: "Type[UserModelType]", role_model: "Optional[Type[RoleModelType]]"
) -> Callable:
    """Get retrieve_user_handler functions for session backends.

    Args:
        user_model: A subclass of a `User` ORM model.
        role_model: A subclass of a `Role` ORM model.
    """

    async def retrieve_user_handler(
        session: Dict[str, Any], connection: "ASGIConnection[Any, Any, Any]"
    ) -> "Optional[UserModelType]":
        """Get a user from a session.

        Args:
            session: Starlite session.
            connection: The ASGI connection.
        """
        async_session_maker = connection.app.state.session_maker_class

        async with async_session_maker() as async_session:
            async with async_session.begin():
                if issubclass(user_model, SQLAlchemyUserRoleMixin) and role_model:
                    repository = SQLAlchemyUserRoleRepository(
                        session=async_session, user_model_type=user_model, role_model_type=role_model
                    )
                else:
                    repository = SQLAlchemyUserRepository(session=async_session, user_model_type=user_model)  # type: ignore[assignment]
                try:
                    user = await repository.get_user(session.get("user_id", ""))
                    if user.is_active and user.is_verified:
                        return user  # type: ignore[return-value]
                except RepositoryNotFoundException:
                    pass
        return None

    return retrieve_user_handler


def get_jwt_retrieve_user_handler(
    user_model: "Type[UserModelType]", role_model: "Optional[Type[RoleModelType]]"
) -> Callable:
    """Get retrieve_user_handler functions for jwt backends.

    Args:
        user_model: A subclass of a `User` ORM model.
        role_model: A subclass of a `Role` ORM model.
    """

    async def retrieve_user_handler(token: "Token", connection: "ASGIConnection[Any, Any, Any]") -> Optional[user_model]:  # type: ignore[valid-type]
        """Get a user from a JWT.

        Args:
            token: Encoded JWT.
            connection: The ASGI connection.
        """
        async_session_maker = connection.app.state.session_maker_class

        async with async_session_maker() as async_session:
            async with async_session.begin():
                if issubclass(user_model, SQLAlchemyUserRoleMixin) and role_model:
                    repository = SQLAlchemyUserRoleRepository(
                        session=async_session, user_model_type=user_model, role_model_type=role_model
                    )
                else:
                    repository = SQLAlchemyUserRepository(session=async_session, user_model_type=user_model)  # type: ignore[assignment]
                try:
                    user = await repository.get_user(token.sub)
                    if user.is_active and user.is_verified:
                        return user
                except RepositoryNotFoundException:
                    pass
        return None

    return retrieve_user_handler



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['starlite_users', 'starlite_users.adapter.sqlalchemy']

package_data = \
{'': ['*']}

install_requires = \
['cryptography', 'passlib', 'python-jose', 'sqlalchemy', 'starlite']

setup_kwargs = {
    'name': 'starlite-users',
    'version': '0.3.0',
    'description': 'Authentication and user management for Starlite',
    'long_description': '# starlite-users\n\nAuthentication, authorization and user management for the Starlite framework\n\n## _This package is not yet production ready._\n\n## Features\n\n- Supports Session, JWT and JWTCookie authentication backends\n- Authorization via role based guards\n- Pre-configured route handlers for:\n  - Authentication\n  - Registration\n  - Verification\n  - Password reset\n  - Administrative user management (read, update, delete)\n  - Administrative role management (read, update, delete)\n  - Assignment/revocation of roles to/from users\n\n## Getting started\n\n### Installation\n\n`pip install starlite-users`\n\n### Documentation\n\n[Read the documentation](https://lonelyvikingmichael.github.io/starlite-users/)\n\nOtherwise [check out the examples](https://github.com/LonelyVikingMichael/starlite-users/tree/main/examples)\n',
    'author': 'Michael Bosch',
    'author_email': 'michael@lonelyviking.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""Modified from https://docs.sqlalchemy.org/en/14/core/custom_types.html#backend-agnostic-guid-type Copyright 2005-2022
SQLAlchemy authors and contributors.
"""
from typing import Any, Union
from uuid import UUID

from sqlalchemy.dialects.postgresql import UUID as PGUUID
from sqlalchemy.types import CHAR, TypeDecorator


class GUID(TypeDecorator):  # pylint: disable=W0223
    """Platform-independent GUID type.

    Uses PostgreSQL's UUID type, otherwise uses CHAR(32), storing as stringified hex values.
    """

    impl = CHAR
    cache_ok = True

    def load_dialect_impl(self, dialect: Any) -> Any:
        """Return a `.TypeEngine` object corresponding to a dialect."""

        if dialect.name == "postgresql":
            return dialect.type_descriptor(PGUUID(as_uuid=True))

        return dialect.type_descriptor(CHAR(36))

    def process_bind_param(self, value: Any, dialect: Any) -> Union[Any, str]:
        """Receive a bound parameter value to be converted."""

        if value is None:
            return value
        if dialect.name == "postgresql":
            return str(value)

        if not isinstance(value, UUID):
            return str(UUID(value))

        return str(value)

    def process_result_value(self, value: Any, dialect: Any) -> Union[Any, UUID]:
        """Receive a result-row column value to be converted."""

        if value is None:
            return value

        if not isinstance(value, UUID):
            return UUID(value)

        return value
